import section, bisect, utils, maxent
import math

class BeamQueue:

    def __init__(self, size):
        
        self.size = size

        self.min_score = None
        self.min_path = None

        self.max_score = None
        self.max_path = None

        self.score_list = []
        self.confidence_list = []
        self.path_list = []


    def add_initial( self, word_list, score ):
        self.min_score = score
        self.min_path = word_list
        
        self.max_score = score
        self.max_path = word_list

        self.score_list.insert (0, score)
        self.confidence_list.insert (0, 0)
        self.path_list.insert (0, word_list)


    def add(self, old_word_list, new_word, score):
        
        # If the beam queue is not full,
        # add the element to the queue
        if len( self.path_list ) < self.size:
            self.__add_to_queue( old_word_list, new_word, score )

        # If the beam is full, and the new element score is greater
        # than or equal to the minimum score in the beam
        elif score >= self.min_score:

            # Remove min element from queue
            self.score_list.pop (0)
            self.path_list.pop (0)

            # Add this element to the queue
            self.__add_to_queue( old_word_list, new_word, score )

        # Make sure the size of the queue is correct
        assert( len(self.path_list) <= self.size )


    # Internal method to add an element to the queue
    def __add_to_queue(self, word_list, new_word, score):

        # Create the new path
        new_path = word_list + [ new_word ]

        # Add new path to the queue
        insertion_point = bisect.bisect (self.score_list, score)
        self.score_list.insert (insertion_point, score)
        self.path_list.insert (insertion_point, new_path)

        # Set the new max
        if score >= self.max_score:
            self.max_path = new_path
            self.max_score = score
        
        self.min_path = self.path_list [0]
        self.min_score = self.score_list [0]


    def items(self):

        for i in range (len (self.score_list)):
            yield self.path_list [i], self.score_list [i]


    def get_best(self):

        assert( len(self.path_list) > 0 )
        return self.max_path


def beam_search(sec, previous_sec, length, beam_size, window_size, model):
     
    bq = BeamQueue( beam_size )

    # Initialize BeamQueue
    initial_word_list = map( lambda x: section.START_WORD, range(window_size - 1) )
    initial_score = 0
    bq.add_initial( initial_word_list, initial_score )

    # Beam search 
    for i in range(length):
        
        new_bq = BeamQueue( beam_size )

        for (path_words, path_score) in bq.items():

            next_fvs = []
            next_words = []

            # Loop through each unique word
            #-----------------------------------------------------------
            # for loop for use with de-stemming...                      
            # for word in model.stem_model.get_words_by_section( sec ): 
            #-----------------------------------------------------------
            for word in sec.unique_word_objs:
                if (word.stem in map (lambda x : x.stem, path_words)):
                    continue
                next_fvs.append( maxent.FeatureVector(path_words, path_words[-window_size+1:], word, False, sec, previous_sec, model ) )
                next_words.append( word )

            path_probs = model.evaluate (next_fvs)
            for next_index, next_score in enumerate( path_probs ):
                if next_score == 0.0:
                    new_path_score = path_score - 100
                else:
                    new_path_score = path_score + math.log( next_score )
                new_bq.add( path_words, next_words[next_index], new_path_score )

        # Set the new beam
        bq = new_bq
        utils.p("    :: " +  ' '.join( [x.word + '/' + x.pos_tag for x in bq.get_best()] ) + " " + str(i)  )

    # Last iteration to end with #END#
    new_bq = BeamQueue(beam_size)
    for (path_words, path_score) in bq.items():
        fv = maxent.FeatureVector(path_words, path_words[-window_size+1:], section.END_WORD, False, sec, previous_sec, model )
        next_score = model.evaluate( [ fv ] )[0]
        if next_score == 0.0:
            new_path_score = path_score - 100
        else:
            new_path_score = path_score + math.log( next_score )
        new_bq.add( path_words, section.END_WORD, new_path_score )
        
    #utils.p(":: " + ' '.join( [x.word + '/' + x.pos_tag for x in new_bq.get_best()] ) + " " + str(i)  )
    for b in range(1, len( new_bq.score_list) ):
        utils.p("    := " + str (b) + ' ' + ' '.join( [x.word + '/' + x.pos_tag for x in new_bq.path_list[-b]] ) + " " + str(new_bq.score_list[-b])  )

    return sec.capitalize_title (new_bq.get_best())
